api_service_name = "RxuniverseApiSvc"

def run_env
  # unsetenv_others: true breaks when calling bundler for unknown reasons ("Not a directory" error)
  %w(BUNDLER_ORIG_BUNDLER_VERSION BUNDLER_ORIG_BUNDLE_BIN_PATH BUNDLER_ORIG_BUNDLE_GEMFILE BUNDLER_ORIG_GEM_HOME BUNDLER_ORIG_GEM_PATH BUNDLER_ORIG_MANPATH BUNDLER_ORIG_PATH BUNDLER_ORIG_RB_USER_INSTALL BUNDLER_ORIG_RUBYLIB BUNDLER_ORIG_RUBYOPT BUNDLER_VERSION BUNDLE_BIN_PATH BUNDLE_GEMFILE GEM_HOME GEM_PATH RUBYLIB RUBYOPT).map do |k|
    [k, nil]
  end.to_h.merge({
                     'Path' => "#{RbConfig::CONFIG['bindir']};#{ENV['Path']}",
                     'TZ' => ENV['TZ']
                 })
end

def simple_extract(package, destination, application_root)
  require 'zip'
  ::Zip.on_exists_proc = true
  ::Zip.validate_entry_sizes = false
  ::Zip::File.open(package) do |zip_file|
    zip_file.each do |entry|
      file_path = File.join(destination, entry.name)
      FileUtils.mkdir_p(File.dirname(file_path)) unless File.exist?(File.dirname(file_path))
      if /vendor/i.match?(file_path) # we don't need to clobber entries in the vendor path - who cares if it's the original rails-6.0.2.gem or the new packaged rails-6.0.2.gem?
        begin
          zip_file.extract(entry, file_path)
        rescue StandardError => err
          safe_msg_dump("Did not clobber file: #{file_path} #{err.message}", application_root)
        end
      else
        zip_file.extract(entry, file_path)
      end
    end
  end
end

def run_command(cmd, application_root, command_desc, working_dir = nil)
  ch_dir = working_dir || application_root
  safe_msg_dump(command_desc, application_root)
  if cmd && cmd != ''
    safe_msg_dump("Running command: #{cmd} in #{ch_dir}", application_root)
    stdout, stderr, status = Open3.capture3(run_env, cmd, chdir: ch_dir)
    if stdout && stdout != ''
      safe_msg_dump("stdout: \n\t#{stdout.gsub("\n", "\n\t")}", application_root)
    end
    if stderr && stderr != ''
      safe_msg_dump("stderr: \n\t#{stderr.gsub("\n", "\n\t")}", application_root)
    end
    raise "#{command_desc} failed (#{status})" unless status.success?
    return stdout
  elsif block_given?
    begin
      safe_msg_dump("Running command: #{command_desc} in #{ch_dir}", application_root)
      ret = nil
      Dir.chdir(ch_dir) do
        ret = yield
      end
      return ret
    rescue StandardError => err
      safe_msg_dump("#{command_desc} failed (#{err})", application_root)
      raise err
    end
  end
  safe_msg_dump("#{command_desc} complete", application_root)
end

def get_file(zip, filename)
  contents = nil
  ::Zip::File.open(zip) do |zip_file|
    if block_given?
      yield zip_file.find_entry(filename)
    else
      contents = zip_file.find_entry(filename).get_input_stream.read
    end
  end
  contents
end

def safe_msg_dump(msg, application_root, mode = 'a')
  begin
    puts msg
    begin
      Dir.chdir(application_root) do
        File.open('./update-log.txt', mode) do |f|
          f.puts "#{Time.now} #{msg}"
          f.flush
        end
      end
    rescue StandardError => dirErr
      Dir.chdir(__dir__) do
        File.open('./update-log.txt', mode) do |f|
          f.puts "#{Time.now} msg"
          f.flush
        end
      end
    rescue StandardError => err
      warn msg
      warn err
    end
  end
end

def update_result_message_file(details, application_root)
  begin
    Dir.chdir(application_root) do
      File.open(File.join(application_root, 'update.result.txt'), 'w') do |f|
        f.puts JSON.generate(details)
        f.flush
      end
    end
  rescue StandardError => err
    safe_msg_dump("Could not update result file (#{err.message})", application_root)
  end
end

def mark_applying(application_root)
  update_result_message_file({applying: true, error: nil, time: Time.now.strftime('%Y-%m-%d %H:%M:%S')}, application_root)
end

def mark_completed(application_root)
  update_result_message_file({applying: false, error: nil, time: Time.now.strftime('%Y-%m-%d %H:%M:%S')}, application_root)
end

def mark_error(exception, application_root)
  update_result_message_file({applying: false, error: exception.message, time: Time.now.strftime('%Y-%m-%d %H:%M:%S'), trace: exception.backtrace}, application_root)
end

task :apply_package, [:package] do |_, args|
  begin
    require 'win32/service'
    require 'zip'
    require 'open3'
    require 'fileutils'
    require 'tmpdir'
    require 'json'
    include Win32
    _, update_package = ARGV
    update_package ||= args[:package]
    application_root = File.expand_path(File.join(__dir__, '..'))
    safe_msg_dump("==============================================================================================================", application_root, 'w')
    safe_msg_dump("application_root: #{application_root}", application_root)
    safe_msg_dump("update_package: #{update_package}", application_root)
    mark_applying(application_root)
    begin
      run_command(nil, application_root, "Stopping service (#{api_service_name})") do
        Service.stop(api_service_name)
      end
    rescue StandardError => e
      safe_msg_dump("Failed to stop service: #{e.message}", application_root)
    end
    run_command(nil, application_root, "Validating update package") do
      get_file(update_package, 'package.json') do |package_detail_file|
        raise "No package.json found in #{update_package}" unless package_detail_file
        package_details = JSON.parse(package_detail_file.get_input_stream.read, symbolize_names: true)
        raise "Expected a package for rx_universe_api, package product id is: #{package_details[:name]} in #{update_package}" unless 'rx_universe_api' == package_details[:name]
      end
    end

    bundle = "\"#{File.join(RbConfig::CONFIG['bindir'], "bundle").tr('/', '\\')}\""
    bundle_env = [
        {cmd: "config --local deployment true", desc: 'Setting bundler deployment mode'},
        {cmd: "config --local system false", desc: 'Setting bundler system mode'},
        {cmd: "config --local path \"vendor/bundle\"", desc: 'Setting bundler path'},
        {cmd: "config --local auto_install true", desc: 'Setting bundler auto-install'},
        {cmd: "config --local allow_offline_install true", desc: 'Setting bundler offline install'},
        {cmd: "config --local disable_shared_gems true", desc: 'Setting bundler shared gems'},
        {cmd: "config --local error_on_stderr true", desc: 'Setting bundler stderr pipe'},
        {cmd: "config --local no-prune false", desc: 'Setting bundler no-prune'},
        {cmd: "config --delete without", desc: 'Setting bundler install groups'},
        {cmd: "config --delete with", desc: 'Setting bundler install groups'},
        {cmd: "env", desc: 'Printing bundler environment info'},
    ].map do |command|
      {cmd: "#{bundle} #{command[:cmd]}", desc: command[:desc]}
    end
    bundle_install = "#{bundle} install --local --gemfile \"#{File.join(application_root, 'Gemfile')}\" --verbose"
    rake = "#{bundle} exec \"#{File.join(RbConfig::CONFIG['bindir'], "rake").tr('/', '\\')}\""
    code_archive = "#{Time.now.strftime("%Y_%m_%d_%H_%M_%S")}.zip"
    rake_backup = "#{rake} backup[#{code_archive}]"
    run_command(rake_backup, application_root, "Backing up code")
    code_archive = File.expand_path(File.join(application_root, "storage", "backups", code_archive))
    safe_msg_dump("Backup archive is #{code_archive}", application_root)

    bundle_env.each do |e|
      run_command(e[:cmd], application_root, e[:desc])
    end

    run_command(nil, application_root, "Clearing existing src folder") do
      remove_dir(File.join(application_root, 'src'))
    end
    run_command(nil, application_root, "Extracting #{update_package} to #{application_root}") do
      simple_extract(update_package, application_root, application_root)
    end
    run_command(bundle_install, application_root, "Bundling dependencies")
    mark_completed(application_root)
    begin
      FileUtils.rm_r(update_package)
    rescue StandardError => err
      # Ignored, no failure to delete the update package constitutes an error in the update process
    end
  rescue StandardError => err
    mark_error(err, application_root)
    safe_msg_dump("@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@@ERR@", application_root)
    safe_msg_dump("Something went wrong, rolling back: #{err.message}\n\t#{err.backtrace.join("\n\t")}", application_root)
    if code_archive && File.exist?(code_archive)
      begin
        run_command(nil, application_root, "Clearing src folder") do
          remove_dir(File.join(application_root, 'src'))
        end
      rescue StandardError => err
        safe_msg_dump("Error clearing app folder: #{err.message}", application_root)
      end
      begin
        run_command(nil, application_root, "Extracting backup package (#{code_archive}") do
          simple_extract(code_archive, application_root, application_root)
        end
        run_command(bundle_install, application_root, "Reverting dependencies")
        safe_msg_dump("Reverted code changes", application_root)
      rescue StandardError => err
        safe_msg_dump("Something went wrong: #{err.message}\n\t#{err.backtrace.join("\n\t")}", application_root)
      end
    end
  ensure
    run_command(nil, application_root, "Starting service (#{api_service_name})") do
      Service.start(api_service_name)
    end
  end
end
